﻿//
// Myriadbits Disk space analyzer 
// Find free space on your NTFS disk
// Copyright (C) 2014 Myriadbits, Jochem Bakker
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// For more information, contact me at: info@myriadbits.com
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.IO;
using Myriadbits.NTFS;

namespace Myriadbits.DiskSpaceAnalyzer
{
	
	/// <summary>
	/// The start form, select a drive, scan the drive
	/// </summary>
	public partial class StartForm : Form
	{
		private Stopwatch sw = new Stopwatch();
		private List<DriveData> m_driveData = new List<DriveData>();
		private bool m_scanningFinished = false;
		private Color m_barColor = Color.FromArgb(40, 0, 190, 50);
		private Color m_barColorBorder = Color.FromArgb(60, 50, 50, 50);
		private UnitTypes m_currentUnit = 0;
		private bool m_fAutoStart = false;
		public List<DriveData> DriveData { get { return m_driveData; } }

		/// <summary>
		/// Load the start form
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void StartForm_Load(object sender, EventArgs e)
		{
			LoadUserSettings();

			this.lblVersion.Text = typeof(AboutForm).Assembly.GetName().Version.ToString();

			lvwDrives.Items.Clear();
			m_driveData.Clear();
			DriveInfo[] allDrives = DriveInfo.GetDrives();
			foreach (DriveInfo d in allDrives)
			{
				string driveName = d.Name.Substring(0, 1);
				string volname = string.IsNullOrEmpty(d.VolumeLabel) ? "Local Disk" : d.VolumeLabel;
				string driveSt = string.Format("{0} ({1})", volname, d.Name);
				ListViewItem lvi = null;
				if (NTFSReader.IsDriveNTFS(driveName))
				{
					lvi = this.lvwDrives.Items.Add(driveSt);
					lvi.Checked = true;
					lvi.Tag = new DriveData(d.Name, driveSt, lvi);
				}
				else
				{
					driveSt = String.Format("{0} - NO NTFS! (this disk cannot be analyzed)", driveSt);
					lvi = this.lvwDrives.Items.Add(driveSt);
					lvi.Checked = false;
					lvi.Tag = null;
				}
				lvi.SubItems.Add(this.m_currentUnit.GetUnitText(d.TotalSize));
				lvi.SubItems.Add(this.m_currentUnit.GetUnitText(d.AvailableFreeSpace));

				if (d.TotalSize != 0)
				{
					double percentage = 100.0 - (d.AvailableFreeSpace * 100) / d.TotalSize;
					lvi.SubItems.Add(percentage.ToString("F01"));
				}
				else
					lvi.SubItems.Add("0");
				lvi.SubItems.Add("<placeholder>"); // Placeholder for progress bar
			}

			if (m_fAutoStart)
				btnStart_Click(null, null);
		}


		/// <summary>
		/// Default constructor
		/// </summary>
		public StartForm(bool autoStart = false)
		{
			m_fAutoStart = autoStart;
			InitializeComponent();
		}

	
		/// <summary>
		/// Draw progress bar
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwDrives_DrawSubItem(object sender, DrawListViewSubItemEventArgs e)
		{
			e.DrawDefault = true;

			// Find the matching drive
			if (e.ColumnIndex == 4)
			{
				DriveData dd = e.Item.Tag as DriveData;
				if (dd != null)
				{
					Rectangle rectFull = e.Bounds;
					e.DrawBackground();

					rectFull.Inflate(-4, -4);

					Rectangle rect;
					rect = new Rectangle(rectFull.Left, rectFull.Top, (int)(dd.Progress * ((double)rectFull.Width)), rectFull.Height);
					e.Graphics.FillRectangle(new SolidBrush(this.m_barColor), rect);	
					e.Graphics.DrawRectangle(new Pen(this.m_barColorBorder, 1), rectFull);

					e.DrawDefault = false;
				}
			}
		}


		/// <summary>
		/// Draw column header
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwDrives_DrawColumnHeader(object sender, DrawListViewColumnHeaderEventArgs e)
		{
			e.DrawDefault = true;
		}


		/// <summary>
		/// Start collecting
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void btnStart_Click(object sender, EventArgs e)
		{
			if (this.m_scanningFinished)
			{
				// We are now the finished button
				this.DialogResult = System.Windows.Forms.DialogResult.OK;
				this.Close();
				return;
			}

			this.lvwDrives.Focus();
			this.btnStart.Enabled = false;
			this.btnExit.Enabled = false;
			
			this.colSize.Text = "Elapsed time";
			this.colFreeSpace.Text = "Estimated time";
			this.colPercentage.Text = "Percentage";
			this.colProgress.Text = "Progress";

			this.m_driveData.Clear();
			foreach(ListViewItem item in this.lvwDrives.Items)
			{
				if (item.Checked)
				{
					if (item.Tag != null)
					{
						DriveData dd = item.Tag as DriveData;
						this.m_driveData.Add(dd);

						dd.Parser = new NTFSReader();
						dd.Parser.Logger += new LogHandler(parser_Logger);
						dd.Parser.ProgressChanged += new Myriadbits.NTFS.ProgressChangedEventHandler(parser_ProgressChanged);
						dd.Parser.BuildTreeFinished += new BuildTreeFinishedEventHandler(parser_BuildTreeFinished);
						dd.Parser.ReadVolumeTree(dd.Drive, 10000, true, true); // TODO Option show hidden files
					}
				}
			}
		}

		/// <summary>
		/// Display logging info
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="message"></param>
		void parser_Logger(object sender, string message)
		{
			this.Invoke((MethodInvoker)delegate
			{
				Debug.WriteLine(message);
			});
		}


		/// <summary>
		/// Show the current progress
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="progress"></param>
		void parser_ProgressChanged(object sender, NTFSParserProgress progress)
		{
			this.Invoke((MethodInvoker)delegate
			{
				// Find the matching drive
				DriveData dd = FindDriveData(sender);
				if (dd != null)
				{
					if (!dd.Time.IsRunning)
					{
						dd.Time.Restart();
					}
					else
					{
						double speed = 1;
						double estimatedTime = 0;

						if (progress.Current > 0.1 * progress.Maximum)
						{
							if (dd.Time.ElapsedMilliseconds != 0)
								speed = ((double)progress.Current) / dd.Time.ElapsedMilliseconds;
							estimatedTime = progress.Maximum / (speed * 1000.0);

							if (dd.TimeEstimation == 0)
								dd.TimeEstimation = estimatedTime; // Estimate that we are a bit faster
							else
								dd.TimeEstimation = 0.95 * dd.TimeEstimation + 0.05 * estimatedTime;
						}

						dd.Progress = 1;
						if (progress.Maximum != 0)
							dd.Progress = ((double)progress.Current) / ((double) progress.Maximum);
						double elapsed = ((double)dd.Time.ElapsedMilliseconds) / 1000.0;

						dd.ListItem.SubItems[1].Text = string.Format("{0:0.0} s", elapsed);
						dd.ListItem.SubItems[2].Text = string.Format("{0:0.0} s", dd.TimeEstimation);
						dd.ListItem.SubItems[3].Text = string.Format("{0:0.}%", dd.Progress * 100.0);
						dd.ListItem.SubItems[4].Text = dd.ListItem.SubItems[3].Text;
					}				
				}
			});
		}


		/// <summary>
		/// The tree has been build
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="mainFolder"></param>
		void parser_BuildTreeFinished(object sender, Folder mainFolder)
		{
			this.Invoke((MethodInvoker)delegate
			{
				// Set this drive data object to finish
				DriveData dd = FindDriveData(sender);
				if (dd != null)
				{
					dd.Finished = true;
					dd.Time.Stop();
				}

				// Check if all drives have finished
				if (this.m_driveData.All(a => a.Finished))
				{
					this.m_scanningFinished = true;
					this.btnExit.Enabled = true;
					this.btnStart.Enabled = true;
					this.btnStart.Text = "Finish";

					// All drives have finished, re-enable ok, finish this dialog
					if (this.chkAutoClose.Checked)
					{
						this.DialogResult = System.Windows.Forms.DialogResult.OK;
						this.Close();
					}
				}
			});
		}


		/// <summary>
		/// Disable start button when no ntfs drive is selected
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void lvwDrives_ItemChecked(object sender, ItemCheckedEventArgs e)
		{
			bool selectionOk = false;
			foreach(ListViewItem item in this.lvwDrives.Items)
			{
				if (item.Checked && item.Tag != null)
				{
					selectionOk = true;
				}
			}
			if (selectionOk)
			{
				this.m_scanningFinished = false;
				this.btnStart.Text = "Start";
			}
			this.btnStart.Enabled = selectionOk;
		}


		/// <summary>
		/// Return the drive data 
		/// </summary>
		/// <param name="sender"></param>
		/// <returns></returns>
		private DriveData FindDriveData(object sender)
		{
			NTFSReader parser = sender as NTFSReader;
			if (parser != null)
			{
				return this.m_driveData.Where(a => a.Drive == parser.Drive).FirstOrDefault();
			}
			return null;
		}

		
		/// <summary>
		/// Load all user settings
		/// </summary>
		private void LoadUserSettings()
		{
			this.m_barColor = DiskSpaceAnalyzer.Properties.Settings.Default.Bar_Color1;
			this.m_currentUnit = (UnitTypes)DiskSpaceAnalyzer.Properties.Settings.Default.Main_Unit;
			this.chkAutoClose.Checked = DiskSpaceAnalyzer.Properties.Settings.Default.AutoCloseScanWindow;
		}


		/// <summary>
		/// Store settings after this box is checked
		/// </summary>
		/// <param name="sender"></param>
		/// <param name="e"></param>
		private void chkAutoClose_CheckedChanged(object sender, EventArgs e)
		{
			DiskSpaceAnalyzer.Properties.Settings.Default.AutoCloseScanWindow = this.chkAutoClose.Checked;
			DiskSpaceAnalyzer.Properties.Settings.Default.Save();
		}
	}


	/// <summary>
	/// Extra class for storing info in the listview tag
	/// </summary>
	public class DriveData
	{
		public DriveData(string drive, string driveDisplayName, ListViewItem lvi)
		{
			this.Drive = drive.Substring(0, 1);
			this.DriveDisplayName = driveDisplayName;
			this.Finished = false;
			this.Time = new Stopwatch();
			this.ListItem = lvi;
			this.TimeEstimation = 0;
			this.Progress = 0;
		}

		public string Drive { get; set; }
		public string DriveDisplayName { get; set; }
		public NTFSReader Parser { get; set; }
		public bool Finished { get; set; }
		public ListViewItem ListItem { get; set; }
		public Stopwatch Time { get; set; }
		public double TimeEstimation { get; set; }
		public double Progress { get; set; }
	};

}
